package ru.dbotthepony.kstarbound.world.`object`

import com.google.common.collect.ImmutableMap
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.reflect.TypeToken
import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.JsonDriven
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.defs.`object`.ObjectOrientation
import ru.dbotthepony.kstarbound.util.set
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.LightCalculator
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2i

abstract class WorldObject(
	val world: World<*, *>,
	val prototype: RegistryObject<ObjectDefinition>,
	val pos: Vector2i,
) : JsonDriven(prototype.file.computeDirectory()) {
	val orientations = prototype.value.orientations
	val validOrientations = orientations.size

	// scriptStorage - json object
	var uniqueId: String? = null
	var interactive = false
	var direction = Direction.LEFT

	init {
		world.objects.add(this)
	}

	var isRemoved = false
		private set

	fun remove() {
		if (isRemoved) return
		check(world.objects.remove(this))
	}

	var orientationIndex = -1
		set(value) {
			if (field != value) {
				field = value
				invalidate()
			}
		}

	val orientation: ObjectOrientation? get() {
		return orientations.getOrNull(orientationIndex)
	}

	override fun defs(): Collection<JsonObject> {
		if (orientationIndex in 0 until validOrientations) {
			return listOf(orientations[orientationIndex].json, prototype.jsonObject)
		} else {
			return listOf(prototype.jsonObject)
		}
	}

	val lightColors: ImmutableMap<String, RGBAColor> by LazyData(listOf("lightColor", "lightColors")) {
		dataValue("lightColor")?.let { ImmutableMap.of("default", colors0.fromJsonTree(it)) }
			?: dataValue("lightColors")?.let { colors1.fromJsonTree(it) }
			?: ImmutableMap.of()
	}

	fun addLights(lightCalculator: LightCalculator, xOffset: Int, yOffset: Int) {
		if ("default" in lightColors) {
			lightCalculator.addPointLight(pos.x - xOffset, pos.y - yOffset, lightColors["default"]!!)
		}
	}

	companion object {
		private val colors1 by lazy { Starbound.gson.getAdapter(TypeToken.getParameterized(ImmutableMap::class.java, String::class.java, RGBAColor::class.java)) as TypeAdapter<ImmutableMap<String, RGBAColor>> }
		private val colors0 by lazy { Starbound.gson.getAdapter(RGBAColor::class.java) }
		private val strings by lazy { Starbound.gson.getAdapter(String::class.java) }
		private val directions by lazy { Starbound.gson.getAdapter(Direction::class.java) }

		fun <W : World<*, *>, T : WorldObject> create(world: W, data: JsonObject, factory: (W, RegistryObject<ObjectDefinition>, Vector2i) -> T): T {
			val tilePosition = Starbound.gson.fromJson(data["tilePosition"], Vector2i::class.java) ?: throw IllegalArgumentException("No tilePosition was present in saved data")
			val name = data["name"]?.asString ?: throw IllegalArgumentException("Missing object name")
			val def = Starbound.worldObjects[name] ?: throw IllegalArgumentException("No such object defined for '$name'")
			val obj = factory(world, def, tilePosition)

			data["direction"]?.let {
				obj.direction = directions.fromJsonTree(it)
			}

			data["orientationIndex"]?.let {
				obj.orientationIndex = it.asInt
			}

			data["interactive"]?.let {
				obj.interactive = it.asBoolean
			}

			data["uniqueId"]?.let {
				if (!it.isJsonNull)
					obj.uniqueId = it.asString
			}

			data["parameters"]?.let {
				for ((k, v) in it.asJsonObject.entrySet()) {
					obj.properties[k] = v
				}
			}

			obj.invalidate()

			return obj
		}
	}
}
